namespace OpenRCT2.Legacy.ObjectExporter.AssemblyInfo

open System.Reflection
open System.Runtime.CompilerServices
open System.Runtime.InteropServices

[<assembly: AssemblyTitle("OpenRCT2.Legacy.ObjectExporter")>]
[<assembly: AssemblyDescription("")>]
[<assembly: AssemblyConfiguration("")>]
[<assembly: AssemblyCompany("")>]
[<assembly: AssemblyProduct("OpenRCT2.Legacy.ObjectExporter")>]
[<assembly: AssemblyCopyright("Copyright © OpenRCT2 2017")>]
[<assembly: AssemblyTrademark("")>]
[<assembly: AssemblyCulture("")>]

[<assembly: ComVisible(false)>]
[<assembly: Guid("77950540-d5fa-4705-a56a-ad047b228bea")>]

[<assembly: AssemblyVersion("1.0.*")>]
[<assembly: AssemblyFileVersion("1.0.0.0")>]

do
    ()
